#include "StdAfx.h"
#include "WinRegionMonitor.h"
#include "WinRegionObserver.h"
#include "ErrorFormatUtils.h"

namespace
{
    void Send(HRGN hRgn, wrm::IWinRegionObserver * pEventObserver)
    {
        DWORD size = GetRegionData(hRgn, 0, 0);
        if (!size)
            throw std::runtime_error(str(boost::format("%1% Fail to get size of region data, %2%")%__FUNCTION__%FormatError(GetLastError())));

        std::vector<char> rgnDataBuf(size);
        RGNDATA * pRgnData = reinterpret_cast<RGNDATA*>(&rgnDataBuf.front());

        if (!GetRegionData(hRgn, size, pRgnData))
            throw std::runtime_error(str(boost::format("%1% Fail to GetRegionData, %2%")%__FUNCTION__%FormatError(GetLastError())));

        pEventObserver->WinEventNotify(pRgnData);    
    }
}
namespace wrm
{
    CWinRegionObserver::CWinRegionObserver(HWND hWnd, IWinRegionObserver * pEventObserver, IErrorObserver* pErrorObserver, DWORD TimeOut)
        : m_window(hWnd)
        , m_timeout(TimeOut)
        , m_regionObserver(pEventObserver)
        , m_errorObserver(pErrorObserver)
        , m_stopping(false)
#pragma warning(push)
#pragma warning(disable:4355) //'this' : used in base member initializer list
        , m_watcherThread(boost::bind(&CWinRegionObserver::NotifyThread, this))
#pragma warning (pop)
    {
        if(!::IsWindow(hWnd))
            throw std::runtime_error(str(boost::format("%1% Invalid window handle")%__FUNCTION__));
        assert(m_regionObserver);
    }

    CWinRegionObserver::~CWinRegionObserver()
    {
        m_stopping = true;
        m_watcherThread.join();
    }

    unsigned int CWinRegionObserver::NotifyThread()
    {
        BEGIN_WEH_EXCEPTION_SAFE_BLOCK
        {
            NotifyZero();
            while (!m_stopping)
            {
                Notify();
                ::Sleep(m_timeout);
            }

            return 0;
        }
        END_WEH_EXCEPTION_SAFE_BLOCK(m_errorObserver);
    }

    void CWinRegionObserver::Notify()
    {
        HDC hDC = ::GetDC(m_window);
        if (!hDC)
            return NotifyZero();

        boost::shared_ptr<void> releaseDCAction(hDC, boost::bind(&ReleaseDC, m_window, _1));

        HRGN hRgn = ::CreateRectRgn(0, 0, 0, 0);
        if(!hRgn)
            throw std::runtime_error(str(boost::format("%1% Fail to CreateRectRgn, %2%")%__FUNCTION__%FormatError(GetLastError())));
        
        boost::shared_ptr<void> hrgnGuard(hRgn, &DeleteObject);
        
        int result = ::GetRandomRgn(hDC, hRgn, SYSRGN);
        if (result <= 0)
            throw std::runtime_error(str(boost::format("%1% Fail to GetRandomRgn, %2%")%__FUNCTION__%FormatError(GetLastError())));

        if(m_oldRgn)
        {
            if(::EqualRgn((HRGN)m_oldRgn.get(), hRgn)) 
                return;
        }

        if (m_stopping)
            return;

        Send(hRgn, m_regionObserver);
        m_oldRgn = hrgnGuard;
    }

    void CWinRegionObserver::NotifyZero()
    {
        HRGN hRgn = CreateRectRgn(0, 0, 0, 0);
        if (!hRgn)
            throw std::runtime_error(str(boost::format("%1% Fail to CreateRectRgn, %2%")%__FUNCTION__%FormatError(GetLastError())));

        boost::shared_ptr<void> hrgnGuard(hRgn, &DeleteObject);

        Send(hRgn, m_regionObserver);
    }

    HWND CWinRegionObserver::GetHWND() 
    { 
        return m_window;
    }
}//weh